/*
 * Author : Gouta Ilyes
 * Subject : audio driver for amp (Win32) / igouta@voila.fr
 * Notes :
 *
 */

#include <windows.h>
#include <stdio.h>
#include <dsound.h>
#include "audio.h"
#include "audioIO.h"

extern "C" char App[];

#define SAFE_RELEASE(a) if (a) {a->Release();a=NULL;}

#define BUFFER_LENGTH           2000 /* 2000 ms */

#define BUFFER_SIZE ((writepos+bufferSize-readpos) % bufferSize)
#define PERCENT_FULL (int) (100.0f*((float)BUFFER_SIZE/bufferSize))

static LPDIRECTSOUND8 dsound=NULL;
static LPDIRECTSOUNDBUFFER dsoundbuffer=NULL;

static DWORD writepos=0,readpos=0;
static int bufferSize,blockAlign;
static int isPlaying=0,running;
static DWORD tID;
static int volume=255; /* vol is in [0,255] */

static HANDLE hPooler;
static HANDLE hWritingMutex;

DWORD WINAPI poolingThread(LPVOID s) {
	int hasMutex=0;
	while (running) {
		if (!hasMutex) { // prevent multiple owns of our mutex
			WaitForSingleObject(hWritingMutex,INFINITE);
			hasMutex=1;
		}
		dsoundbuffer->GetCurrentPosition(&readpos,NULL);
#ifdef __DEBUG__
		printf("\rbuffer size : %d%% ",PERCENT_FULL);
#endif
		if (PERCENT_FULL<70) { // remember... we'll sleep for 20 ms
			hasMutex=0;
			ReleaseMutex(hWritingMutex);
		}
		Sleep(20); // sleep for 20 ms
	}
	return 0;
}

int audioOpen(int freq,int stereo,int vol) {
	DSBUFFERDESC bufferdesc;
	WAVEFORMATEX format;
	
	hWritingMutex=CreateMutex(NULL,FALSE,NULL);
	if (!hWritingMutex) {
		MessageBox(NULL,"Failed creating mutex !",App,MB_ICONSTOP);
		return 0;
	}

	ZeroMemory(&format,sizeof(format));
	format.cbSize=sizeof(format);
	format.wFormatTag=WAVE_FORMAT_PCM;
	format.nChannels=(stereo?2:1);
	format.nSamplesPerSec=freq;
	format.nBlockAlign=format.nChannels*2;
	format.wBitsPerSample=16;
	format.nAvgBytesPerSec=format.nBlockAlign*format.nSamplesPerSec;

	ZeroMemory(&bufferdesc,sizeof(bufferdesc));
	bufferdesc.dwSize=sizeof(bufferdesc);
	bufferdesc.dwFlags=DSBCAPS_CTRLVOLUME|DSBCAPS_GLOBALFOCUS|DSBCAPS_GETCURRENTPOSITION2;
	bufferdesc.dwBufferBytes=(int)(((float)BUFFER_LENGTH/1000)*format.nAvgBytesPerSec);
	bufferdesc.dwBufferBytes-=(bufferdesc.dwBufferBytes%format.nBlockAlign);
	bufferdesc.lpwfxFormat=&format;
	bufferdesc.guid3DAlgorithm=GUID_NULL;

	bufferSize=bufferdesc.dwBufferBytes;
	blockAlign=format.nBlockAlign;
	
	if (DirectSoundCreate8(NULL,&dsound,NULL)!=DS_OK) return 0;
	if (dsound->SetCooperativeLevel(GetDesktopWindow(),DSSCL_NORMAL)!=DS_OK) {SAFE_RELEASE(dsound);return 0;}
	if (dsound->CreateSoundBuffer(&bufferdesc,&dsoundbuffer,NULL)!=DS_OK) {SAFE_RELEASE(dsound);return 0;}

	void *bytes1,*bytes2;DWORD size1,size2;
	dsoundbuffer->Lock(0,0,&bytes1,&size1,&bytes2,&size2,DSBLOCK_ENTIREBUFFER);
	ZeroMemory(bytes1,size1);
	dsoundbuffer->Unlock(bytes1,size1,bytes2,size2);

	audioSetVolume(volume);

	isPlaying=0;
	return 1;
}

void audioClose() {
	if (hPooler) {
		running=0;
		if (WaitForSingleObject(hPooler,200)==WAIT_TIMEOUT) {
			MessageBox(NULL,"Failed asking thread to die !",App,MB_ICONEXCLAMATION);
			TerminateThread(hPooler,0);
		}
		CloseHandle(hPooler);
		hPooler=NULL;
	}
	if (hWritingMutex) {
		CloseHandle(hWritingMutex);
		hWritingMutex=NULL;
	}
	if (dsoundbuffer) {
		dsoundbuffer->Stop();
		SAFE_RELEASE(dsoundbuffer);
	}
	if (dsound) SAFE_RELEASE(dsound);
	isPlaying=0;
}

int audioWrite(char* data,int len) {
	void *bytes1,*bytes2;
	DWORD size1,size2;

	WaitForSingleObject(hWritingMutex,INFINITE); // request access to the sound buffer
	
	dsoundbuffer->Lock(writepos,len,&bytes1,&size1,&bytes2,&size2,0);
	if (!bytes2) CopyMemory(bytes1,data,len);
	else {
		CopyMemory(bytes1,data,size1);
		CopyMemory(bytes2,data+size1,size2);
	}
	dsoundbuffer->Unlock(bytes1,size1,bytes2,size2);
	writepos+=len;
	writepos%=bufferSize;
	
	ReleaseMutex(hWritingMutex);

	if (!isPlaying) {
		running=1;isPlaying=1;
		hPooler=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) poolingThread,NULL,0,&tID);
		SetThreadPriority(hPooler,THREAD_PRIORITY_ABOVE_NORMAL); // must be a higher priority !!!
		dsoundbuffer->Play(0,0,DSBPLAY_LOOPING);
	}

	return len;
}

void audioFlush() {}

void audioSetVolume(int vol) {
	if (!dsoundbuffer) return;
	dsoundbuffer->SetVolume((LONG) ((volume=vol)*(DSBVOLUME_MAX-DSBVOLUME_MIN)/256.0f+DSBVOLUME_MIN));
}

int audioGetVolume() {
	if (!dsoundbuffer) return 0;
	dsoundbuffer->GetVolume((LONG*)&volume);
	volume=(int) (256.0f*(volume-DSBVOLUME_MIN)/(DSBVOLUME_MAX-DSBVOLUME_MIN));
	return volume;
}